@@ -11,8 +11,6 @@
 #include <errno.h>
 #include <string.h>
 #include <limits.h>
-#include <pwd.h>
-#include <grp.h>
 #include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
@@ -25,6 +23,7 @@
 #include <sys/vfs.h>
 #include <stdint.h>
 #include <sys/utsname.h>
+#include <sys/syscall.h>
 
 #include "pot.h"
 #include "bylabel.h"
@@ -69,7 +68,17 @@ int nfs_fstype(char *type)
  */
 int meta_qf_fstype(char *type)
 {
-	return !strcmp(type, MNTTYPE_OCFS2);
+	return !strcmp(type, MNTTYPE_OCFS2) || !strcmp(type, MNTTYPE_TMPFS) ||
+	       !strcmp(type, MNTTYPE_BCACHEFS);
+}
+
+/*
+ *	Check whether the filesystem is not using block device as a backing
+ */
+int nodev_fstype(char *type)
+{
+	return !strcmp(type, MNTTYPE_TMPFS) ||
+	       !strcmp(type, MNTTYPE_BCACHEFS) || nfs_fstype(type);
 }
 
 /*
@@ -80,183 +89,6 @@ char *type2name(int type)
 	return extensions[type];
 }
 
-/*
- *	Project quota handling
- */
-#define PROJECT_FILE "/etc/projid"
-#define MAX_LINE_LEN 1024
-
-static FILE *project_file;
-
-/* Rewind /etc/projid to the beginning */
-void setprent(void)
-{
-	if (project_file)
-		fclose(project_file);
-	project_file = fopen(PROJECT_FILE, "r");
-}
-
-/* Close /etc/projid file */
-void endprent(void)
-{
-	if (project_file) {
-		fclose(project_file);
-		project_file = NULL;
-	}
-}
-
-/* Get next entry in /etc/projid */
-struct fs_project *getprent(void)
-{
-	static struct fs_project p;
-	static char linebuf[MAX_LINE_LEN];
-	char *idstart, *idend;
-
-	if (!project_file)
-		return NULL;
-	while (fgets(linebuf, MAX_LINE_LEN, project_file)) {
-		/* Line too long? */
-		if (linebuf[strlen(linebuf) - 1] != '\n')
-			continue;
-		/* Skip comments */
-		if (linebuf[0] == '#')
-			continue;
-		idstart = strchr(linebuf, ':');
-		/* Skip invalid lines... We follow what xfs_quota does */
-		if (!idstart)
-			continue;
-		*idstart = 0;
-		idstart++;
-		/*
-		 * Colon can separate name from something else - follow what
-		 * xfs_quota does
-		 */
-		idend = strchr(idstart, ':');
-		if (idend)
-			*idend = 0;
-		p.pr_name = linebuf;
-		p.pr_id = strtoul(idstart, NULL, 10);
-		return &p;
-	}
-	return NULL;
-}
-
-static struct fs_project *get_project_by_name(char *name)
-{
-	struct fs_project *p;
-
-	setprent();
-	while ((p = getprent())) {
-		if (!strcmp(name, p->pr_name))
-			break;
-	}
-	endprent();
-
-	return p;
-}
-
-static struct fs_project *get_project_by_id(qid_t id)
-{
-	struct fs_project *p;
-
-	setprent();
-	while ((p = getprent())) {
-		if (p->pr_id == id)
-			break;
-	}
-	endprent();
-
-	return p;
-}
-
-/*
- *	Convert name to uid
- */
-uid_t user2uid(char *name, int flag, int *err)
-{
-	struct passwd *entry;
-	uid_t ret;
-	char *errch;
-
-	if (err)
-		*err = 0;
-	if (!flag) {
-		ret = strtoul(name, &errch, 0);
-		if (!*errch)		/* Is name number - we got directly uid? */
-			return ret;
-	}
-	if (!(entry = getpwnam(name))) {
-		if (!err) {
-			errstr(_("user %s does not exist.\n"), name);
-			exit(1);
-		}
-		else {
-			*err = -1;
-			return 0;
-		}
-	}
-	return entry->pw_uid;
-}
-
-/*
- *	Convert group name to gid
- */
-gid_t group2gid(char *name, int flag, int *err)
-{
-	struct group *entry;
-	gid_t ret;
-	char *errch;
-
-	if (err)
-		*err = 0;
-	if (!flag) {
-		ret = strtoul(name, &errch, 0);
-		if (!*errch)		/* Is name number - we got directly gid? */
-			return ret;
-	}
-	if (!(entry = getgrnam(name))) {
-		if (!err) {
-			errstr(_("group %s does not exist.\n"), name);
-			exit(1);
-		}
-		else {
-			*err = -1;
-			return 0;
-		}
-	}
-	return entry->gr_gid;
-}
-
-/*
- *	Convert project name to project id
- */
-qid_t project2pid(char *name, int flag, int *err)
-{
-	qid_t ret;
-	char *errch;
-	struct fs_project *p;
-
-	if (err)
-		*err = 0;
-	if (!flag) {
-		ret = strtoul(name, &errch, 0);
-		if (!*errch)		/* Is name number - we got directly pid? */
-			return ret;
-	}
-	p = get_project_by_name(name);
-	if (!p) {
-		if (!err) {
-			errstr(_("project %s does not exist.\n"), name);
-			exit(1);
-		}
-		else {
-			*err = -1;
-			return 0;
-		}
-	}
-	return p->pr_id;
-}
-
 /*
  *	Convert name to id
  */
@@ -270,54 +102,6 @@ int name2id(char *name, int qtype, int flag, int *err)
 		return project2pid(name, flag, err);
 }
 
-/*
- *	Convert uid to name
- */
-int uid2user(uid_t id, char *buf)
-{
-	struct passwd *entry;
-
-	if (!(entry = getpwuid(id))) {
-		snprintf(buf, MAXNAMELEN, "#%u", (uint) id);
-		return 1;
-	}
-	else
-		sstrncpy(buf, entry->pw_name, MAXNAMELEN);
-	return 0;
-}
-
-/*
- *	Convert gid to name
- */
-int gid2group(gid_t id, char *buf)
-{
-	struct group *entry;
-
-	if (!(entry = getgrgid(id))) {
-		snprintf(buf, MAXNAMELEN, "#%u", (uint) id);
-		return 1;
-	}
-	else
-		sstrncpy(buf, entry->gr_name, MAXNAMELEN);
-	return 0;
-}
-
-/*
- *	Convert project id to name
- */
-int pid2project(qid_t id, char *buf)
-{
-	struct fs_project *p;
-
-	if (!(p = get_project_by_id(id))) {
-		snprintf(buf, MAXNAMELEN, "#%u", (uint) id);
-		return 1;
-	}
-	else
-		sstrncpy(buf, p->pr_name, MAXNAMELEN);
-	return 0;
-}
-
 /*
  *	Convert id to user/groupname
  */
@@ -331,33 +115,6 @@ int id2name(int id, int qtype, char *buf)
 		return pid2project(id, buf);
 }
 
-/*
- *	Parse /etc/nsswitch.conf and return type of default passwd handling
- */
-int passwd_handling(void)
-{
-	FILE *f;
-	char buf[1024], *colpos, *spcpos;
-	int ret = PASSWD_FILES;
-
-	if (!(f = fopen("/etc/nsswitch.conf", "r")))
-		return PASSWD_FILES;	/* Can't open nsswitch.conf - fallback on compatible mode */
-	while (fgets(buf, sizeof(buf), f)) {
-		if (strncmp(buf, "passwd:", 7))	/* Not passwd entry? */
-			continue;
-		for (colpos = buf+7; isspace(*colpos); colpos++);
-		if (!*colpos)	/* Not found any type of handling? */
-			break;
-		for (spcpos = colpos; !isspace(*spcpos) && *spcpos; spcpos++);
-		*spcpos = 0;
-		if (!strcmp(colpos, "db") || !strcmp(colpos, "nis") || !strcmp(colpos, "nis+"))
-			ret = PASSWD_DB;
-		break;
-	}
-	fclose(f);
-	return ret;
-}
-
 /*
  *	Convert quota format name to number
  */
@@ -670,6 +427,36 @@ const char *str2number(const char *string, qsize_t *inodes)
 	return NULL;
 }
 
+/*
+ * Wrappers for quotactl syscalls
+ */
+#ifdef SYS_quotactl_fd
+int do_quotactl(int cmd, const char *dev, const char *mnt, int id, caddr_t addr)
+{
+	int ret = -EINVAL;
+
+	if (mnt && !dev) {
+		int fd = open(mnt, O_DIRECTORY | O_PATH);
+
+		if (fd < 0) {
+			errstr(_("Unable to get a filedescriptor from mountpoint: %s\n"), mnt);
+			return fd;
+		}
+
+		ret = syscall(SYS_quotactl_fd, fd, cmd, id, addr);
+		close(fd);
+		return ret;
+	}
+
+	return quotactl(cmd, dev, id, addr);
+}
+#else
+int do_quotactl(int cmd, const char *dev, const char *mnt, int id, caddr_t addr)
+{
+	return quotactl(cmd, dev, id, addr);
+}
+#endif
+
 /*
  *	Wrappers for mount options processing functions
  */
@@ -685,7 +472,7 @@ static int hasxfsquota(const char *dev, struct mntent *mnt, int type, int flags)
 		return QF_XFS;
 
 	memset(&info, 0, sizeof(struct xfs_mem_dqinfo));
-	if (!quotactl(QCMD(Q_XFS_GETQSTAT, type), dev, 0, (void *)&info)) {
+	if (!do_quotactl(QCMD(Q_XFS_GETQSTAT, type), dev, mnt->mnt_dir, 0, (void *)&info)) {
 #ifdef XFS_ROOTHACK
 		int sbflags = (info.qs_flags & 0xff00) >> 8;
 #endif /* XFS_ROOTHACK */
@@ -717,10 +504,18 @@ static int hasxfsquota(const char *dev, struct mntent *mnt, int type, int flags)
 
 static int hasvfsmetaquota(const char *dev, struct mntent *mnt, int type, int flags)
 {
-	uint32_t fmt;
+	struct if_dqinfo info;
 
-	if (!quotactl(QCMD(Q_GETFMT, type), dev, 0, (void *)&fmt))
+	/*
+	 * Some filesystems such as bcachefs do support QF_META format but
+	 * they don't use generic infrastructure for tracking quotas and
+	 * hence they don't set quota format. Q_GETINFO is a reliable way to
+	 * detect both such filesystems and all other filesystems using QF_META
+	 * format.
+	 */
+	if (!do_quotactl(QCMD(Q_GETINFO, type), dev, mnt->mnt_dir, 0, (void *)&info))
 		return QF_META;
+
 	return QF_ERROR;
 }
 
@@ -785,8 +580,15 @@ static int hasquota(const char *dev, struct mntent *mnt, int type, int flags)
 	    !strcmp(mnt->mnt_type, MNTTYPE_XFS) ||
 	    !strcmp(mnt->mnt_type, MNTTYPE_EXFS))
 		return hasxfsquota(dev, mnt, type, flags);
+
 	if (!strcmp(mnt->mnt_type, MNTTYPE_OCFS2))
 		return hasvfsmetaquota(dev, mnt, type, flags);
+
+	/* tmpfs has no device, bcachefs has possibly many devices. Pass null
+	 * here so quotactl_fd() is called */
+	if (!strcmp(mnt->mnt_type, MNTTYPE_TMPFS) ||
+	    !strcmp(mnt->mnt_type, MNTTYPE_BCACHEFS))
+		return hasvfsmetaquota(NULL, mnt, type, flags);
 	/*
 	 * For ext4 we check whether it has quota in system files and if not,
 	 * we fall back on checking standard quotas. Furthermore we cannot use
@@ -796,7 +598,7 @@ static int hasquota(const char *dev, struct mntent *mnt, int type, int flags)
 	if (!strcmp(mnt->mnt_type, MNTTYPE_EXT4) || !strcmp(mnt->mnt_type, MNTTYPE_F2FS)) {
 		struct if_dqinfo kinfo;
 
-		if (quotactl(QCMD(Q_GETINFO, type), dev, 0, (void *)&kinfo) == 0) {
+		if (do_quotactl(QCMD(Q_GETINFO, type), dev, mnt->mnt_dir, 0, (void *)&kinfo) == 0) {
 			if (kinfo.dqi_flags & DQF_SYS_FILE)
 				return QF_META;
 		}
@@ -1069,11 +871,11 @@ void init_kernel_interface(void)
 		else {
 			fs_quota_stat_t dummy;
 
-			if (!quotactl(QCMD(Q_XGETQSTAT, 0), "/dev/root", 0, (void *)&dummy) ||
+			if (!do_quotactl(QCMD(Q_XGETQSTAT, 0), "/dev/root", NULL, 0, (void *)&dummy) ||
 			    (errno != EINVAL && errno != ENOSYS))
 				kernel_qfmt[kernel_qfmt_num++] = QF_XFS;
 		}
-		if (quotactl(QCMD(Q_V2_GETSTATS, 0), NULL, 0, (void *)&v2_stats) >= 0) {
+		if (do_quotactl(QCMD(Q_V2_GETSTATS, 0), NULL, NULL, 0, (void *)&v2_stats) >= 0) {
 			kernel_qfmt[kernel_qfmt_num++] = QF_VFSV0;
 			kernel_iface = IFACE_VFSV0;
 		}
@@ -1085,9 +887,9 @@ void init_kernel_interface(void)
 			int err_quota = 0;
  			char tmp[1024];         /* Just temporary buffer */
 
-			if (quotactl(QCMD(Q_V1_GETSTATS, 0), NULL, 0, tmp))
+			if (do_quotactl(QCMD(Q_V1_GETSTATS, 0), NULL, NULL, 0, tmp))
 				err_stat = errno;
-			if (quotactl(QCMD(Q_V1_GETQUOTA, 0), "/dev/null", 0, tmp))
+			if (do_quotactl(QCMD(Q_V1_GETQUOTA, 0), "/dev/null", NULL, 0, tmp))
 				err_quota = errno;
 
 			/* On a RedHat 2.4.2-2 	we expect 0, EINVAL
@@ -1127,7 +929,7 @@ static int v1_kern_quota_on(const char *dev, int type)
 	char tmp[1024];		/* Just temporary buffer */
 	qid_t id = (type == USRQUOTA) ? getuid() : getgid();
 
-	if (!quotactl(QCMD(Q_V1_GETQUOTA, type), dev, id, tmp))	/* OK? */
+	if (!do_quotactl(QCMD(Q_V1_GETQUOTA, type), dev, NULL, id, tmp))	/* OK? */
 		return 1;
 	return 0;
 }
@@ -1138,7 +940,7 @@ static int v2_kern_quota_on(const char *dev, int type)
 	char tmp[1024];		/* Just temporary buffer */
 	qid_t id = (type == USRQUOTA) ? getuid() : getgid();
 
-	if (!quotactl(QCMD(Q_V2_GETQUOTA, type), dev, id, tmp))	/* OK? */
+	if (!do_quotactl(QCMD(Q_V2_GETQUOTA, type), dev, NULL, id, tmp))	/* OK? */
 		return 1;
 	return 0;
 }
@@ -1155,7 +957,7 @@ int kern_quota_state_xfs(const char *dev, int type)
 {
 	struct xfs_mem_dqinfo info;
 
-	if (!quotactl(QCMD(Q_XFS_GETQSTAT, type), dev, 0, (void *)&info)) {
+	if (!do_quotactl(QCMD(Q_XFS_GETQSTAT, type), dev, NULL, 0, (void *)&info)) {
 		if (type == USRQUOTA) {
 			return !!(info.qs_flags & XFS_QUOTA_UDQ_ACCT) +
 			       !!(info.qs_flags & XFS_QUOTA_UDQ_ENFD);
@@ -1199,8 +1001,8 @@ int kern_quota_on(struct mount_entry *mnt, int type, int fmt)
 	if (kernel_iface == IFACE_GENERIC) {
 		int actfmt;
 
-		if (quotactl(QCMD(Q_GETFMT, type), mnt->me_devname, 0,
-			     (void *)&actfmt) >= 0) {
+		if (do_quotactl(QCMD(Q_GETFMT, type), mnt->me_devname,
+				mnt->me_dir, 0, (void *)&actfmt) >= 0) {
 			actfmt = kern2utilfmt(actfmt);
 			if (actfmt >= 0)
 				return actfmt;
@@ -1223,7 +1025,7 @@ int kern_quota_on(struct mount_entry *mnt, int type, int fmt)
  */
 
 struct searched_dir {
-	int sd_dir;		/* Is searched dir mountpoint or in fact device? */
+	int sd_isdir;		/* Is searched dir mountpoint or in fact device? */
 	dev_t sd_dev;		/* Device mountpoint lies on */
 	ino_t sd_ino;		/* Inode number of mountpoint */
 	const char *sd_name;	/* Name of given dir/device */
@@ -1353,7 +1155,7 @@ alloc:
 			continue;
 		}
 
-		if (!nfs_fstype(mnt->mnt_type)) {
+		if (!nodev_fstype(mnt->mnt_type)) {
 			if (stat(devname, &st) < 0) {	/* Can't stat mounted device? */
 				errstr(_("Cannot stat() mounted device %s: %s\n"), devname, strerror(errno));
 				free((char *)devname);
@@ -1367,15 +1169,16 @@ alloc:
 			dev = st.st_rdev;
 			for (i = 0; i < mnt_entries_cnt && mnt_entries[i].me_dev != dev; i++);
 		}
-		/* Cope with network filesystems or new mountpoint */
-		if (nfs_fstype(mnt->mnt_type) || i == mnt_entries_cnt) {
+
+		/* Cope with filesystems without a block device or new mountpoint */
+		if (nodev_fstype(mnt->mnt_type) || i == mnt_entries_cnt) {
 			if (stat(mnt->mnt_dir, &st) < 0) {	/* Can't stat mountpoint? We have better ignore it... */
 				errstr(_("Cannot stat() mountpoint %s: %s\n"), mnt->mnt_dir, strerror(errno));
 				free((char *)devname);
 				continue;
 			}
-			if (nfs_fstype(mnt->mnt_type)) {
-				/* For network filesystems we must get device from root */
+			if (nodev_fstype(mnt->mnt_type)) {
+				/* For filesystems without block device we must get device from root */
 				dev = st.st_dev;
 				if (!(flags & MS_NFS_ALL)) {
 					for (i = 0; i < mnt_entries_cnt && mnt_entries[i].me_dev != dev; i++);
@@ -1454,7 +1257,7 @@ static int process_dirs(int dcnt, char **dirs, int flags)
 					errstr(_("Cannot stat() given mountpoint %s: %s\nSkipping...\n"), dirs[i], strerror(errno));
 					continue;
 				}
-			check_dirs[check_dirs_cnt].sd_dir = S_ISDIR(st.st_mode);
+			check_dirs[check_dirs_cnt].sd_isdir = S_ISDIR(st.st_mode);
 			if (S_ISDIR(st.st_mode)) {
 				const char *realmnt = dirs[i];
 
@@ -1538,7 +1341,7 @@ restart:
 		return 0;
 	sd = check_dirs + act_checked;
 	for (i = 0; i < mnt_entries_cnt; i++) {
-		if (sd->sd_dir) {
+		if (sd->sd_isdir) {
 			if (sd->sd_dev == mnt_entries[i].me_dev && sd->sd_ino == mnt_entries[i].me_ino)
 				break;
 		}
