@@ -13,7 +13,6 @@
 #include "quota.h"
 #include "common.h"
 
-#define MAXNAMELEN 64		/* Maximal length of user/group name */
 #define MAXTIMELEN 40		/* Maximal length of time string */
 #define MAXNUMLEN 32		/* Maximal length of number */
 
@@ -47,11 +46,6 @@ struct mount_entry {
 	int me_qfmt[MAXQUOTAS];	/* Detected quota formats */
 };
 
-struct fs_project {
-	qid_t pr_id;
-	char *pr_name;
-};
-
 /*
  *	Exported functions
  */
@@ -63,45 +57,12 @@ int meta_qf_fstype(char *type);
 /* Convert quota type to written form */
 char *type2name(int);
 
-/* Rewind /etc/projid to the beginning */
-void setprent(void);
-
-/* Close /etc/projid file */
-void endprent(void);
-
-/* Get next entry in /etc/projid */
-struct fs_project *getprent(void);
-
-/* Convert username to uid */
-uid_t user2uid(char *, int flag, int *err);
-
-/* Convert groupname to gid */
-gid_t group2gid(char *, int flag, int *err);
-
-/* Convert project name to project id */
-qid_t project2pid(char *name, int flag, int *err);
-
 /* Convert user/groupname to id */
 int name2id(char *name, int qtype, int flag, int *err);
 
-/* Convert uid to username */
-int uid2user(uid_t, char *);
-
-/* Convert gid to groupname */
-int gid2group(gid_t, char *);
-
-/* Convert project id to name */
-int pid2project(qid_t, char *);
-
 /* Convert id to user/group name */
 int id2name(int id, int qtype, char *buf);
 
-/* Possible default passwd handling */
-#define PASSWD_FILES 0
-#define PASSWD_DB 1
-/* Parse /etc/nsswitch.conf and return type of default passwd handling */
-int passwd_handling(void);
-
 /* Convert quota format name to number */
 int name2fmt(char *str);
 
@@ -206,6 +167,9 @@ void end_mounts_scan(void);
 /* Parse kernel version and return 1 if ext4 supports quota feature */
 int ext4_supports_quota_feature(void);
 
+/* Wrapper around quota syscalls, either call quotactl or quotactl_fd */
+int do_quotactl(int cmd, const char *dev, const char *mnt, int id, caddr_t addr);
+
 /* Quota output formats */
 #define QOF_ERROR	-1
 #define QOF_DEFAULT	0
